// MyOpenArchive.cpp: COpenArchive NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyOpenArchive.h"
#include "MainAr.h"
#include "SplitCmdLine.h"
#include <time.h>

#include "Windows/PropVariant.h"
#include "Windows/PropVariantConversions.h"
//#include "Interface/PropID.h"
#include "../../UI/Common/OpenArchive.h"

static COpenArchive* pOATail;

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

COpenArchive::COpenArchive()
{
	m_pOANext = NULL;
	m_pOAPrev = pOATail;
	if (pOATail)
		pOATail->m_pOANext = this;
	pOATail = this;
}

COpenArchive::~COpenArchive()
{
	if (m_pOAPrev)
		m_pOAPrev->m_pOANext = m_pOANext;
	if (m_pOANext)
		m_pOANext->m_pOAPrev = m_pOAPrev;
	if (pOATail == this)
		pOATail = pOATail->m_pOAPrev;

	if (m_pWildcardCensor)
		delete m_pWildcardCensor;
}

// ỸIuWFNg폜
void COpenArchive::RemoveAll()
{
	while (pOATail)
		delete pOATail;
}

// IuWFNg̃|C^
COpenArchive* COpenArchive::FindObject(HARC hArc)
{
	for (COpenArchive* pOA = pOATail; pOA; pOA = pOA->m_pOAPrev)
	{
		if (pOA == (COpenArchive*)hArc)
			return pOA;
	}
	return NULL;
}

BOOL COpenArchive::Open(LPCSTR lpFileName, DWORD dwMode)
{
	m_pWildcardCensor = NULL;
	m_dwMode = dwMode;
	m_bSearchMode = FALSE;
	HANDLE hFind = ::FindFirstFile(GetSystemString(lpFileName, CP_OEMCP), &m_fdArchive);
	if (hFind == INVALID_HANDLE_VALUE)
		return FALSE;
	::FindClose(hFind);

    CArchiverInfo anArchiverInfo;
    COpenCallbackImp *openCallbackSpec = new COpenCallbackImp;
    CMyComPtr<IArchiveOpenCallback> openCallback = openCallbackSpec;
	int subExtIndex;

	if (OpenArchive(GetSystemString(lpFileName, CP_OEMCP), &m_anArchiveHandler, anArchiverInfo, subExtIndex, openCallback) != S_OK)
		return FALSE;
	if (m_anArchiveHandler->GetNumberOfItems(&m_aNumItems) != S_OK)
		return FALSE;
	if (anArchiverInfo.Name.CompareNoCase(L"7z") == 0)
		m_nArchiveType = ARCHIVETYPE_7Z;
	else if (anArchiverInfo.Name.CompareNoCase(L"zip") == 0)
		m_nArchiveType = ARCHIVETYPE_ZIP;
	return TRUE;
}

int COpenArchive::FindFirst(LPCSTR lpszWildName, INDIVIDUALINFO *lpSubInfo)
{
	CSplitCmdLine scl;
	if (!scl.Split(lpszWildName))
		return -1;
	if (m_pWildcardCensor)
		delete m_pWildcardCensor;
	m_pWildcardCensor = new NWildcard::CCensor;
	for (int i = 1; i < scl.m_argc; ++i)
	{
		UString uString = MultiByteToUnicodeString(scl.m_argv[i], CP_OEMCP);
		m_pWildcardCensor->AddItem(uString, true, true, DoesNameContainWildCard(uString));
	}
	if (scl.m_argc == 1)
		m_pWildcardCensor->AddItem(L"*", true, true, true);
	m_aItemPos = 0;
	m_ui64ArcOriginalSize = 0;
	m_ui64ArcCompressedSize = 0;
	m_bSearchMode = TRUE;
	m_bSolid = FALSE;
	if (m_nArchiveType == ARCHIVETYPE_7Z)
	{
		NWindows::NCOM::CPropVariant aPropVariant;
		BOOL bFlag = 0;
		for (int i = 0; i < m_aNumItems; ++i)
		{
			m_anArchiveHandler->GetProperty(i, kpidBlock, &aPropVariant);
			if (aPropVariant.vt == 0)
				continue;
			if (aPropVariant.ulVal > 0)
				break;
			if (bFlag == TRUE)
			{
				m_bSolid = TRUE;
				break;
			}
			bFlag = TRUE;
		}
	}
	return FindNext(lpSubInfo);
}

int COpenArchive::FindNext(INDIVIDUALINFO *lpSubInfo)
{
	if (!m_bSearchMode)
		return ERROR_NOT_SEARCH_MODE;

	NWindows::NCOM::CPropVariant aPropVariant;
	while (TRUE)
	{
		if (m_aItemPos == m_aNumItems)
		{
			m_bSearchMode = FALSE;
			return -1;
		}
		m_anArchiveHandler->GetProperty(m_aItemPos, kpidPath, &aPropVariant);
		m_aFileName = GetOemString(aPropVariant.bstrVal);
		if (m_pWildcardCensor->CheckName(aPropVariant.bstrVal))
			break;
		++m_aItemPos;
	}
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidIsFolder, &aPropVariant);
	if (aPropVariant.boolVal)
	{
		m_aFileName.TrimRight('\\');
		m_aFileName += "\\";
	}
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidSize, &aPropVariant);
	m_ui64OriginalSize = ConvertPropVariantToUINT64(aPropVariant);
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidPackedSize, &aPropVariant);
	m_ui64CompressedSize = m_bSolid ? 0 : ConvertPropVariantToUINT64(aPropVariant);
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidCRC, &aPropVariant);
	m_dwCRC = aPropVariant.vt ? aPropVariant.ulVal : 0;
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidLastWriteTime, &aPropVariant);
	m_ftLastWriteTime = aPropVariant.filetime;
	m_anArchiveHandler->GetProperty(m_aItemPos, kpidAttributes, &aPropVariant);
	m_dwAttribute = aPropVariant.ulVal;
	
	if (m_nArchiveType == ARCHIVETYPE_ZIP)
	{
		m_anArchiveHandler->GetProperty(m_aItemPos, kpidEncrypted, &aPropVariant);
		if (aPropVariant.boolVal)
			m_dwAttribute |= FA_ENCRYPTED;
		m_anArchiveHandler->GetProperty(m_aItemPos, kpidMethod, &aPropVariant);
		m_aMode = GetOemString(aPropVariant.bstrVal);

//		m_anArchiveHandler->GetProperty(m_aItemPos, kpidHostOS, &aPropVariant);
//		m_anArchiveHandler->GetProperty(m_aItemPos, kpidComment, &aPropVariant);
	}
	else
	{
		m_anArchiveHandler->GetProperty(m_aItemPos, kpidMethod, &aPropVariant);
		m_aMode = aPropVariant.vt ? GetOemString(aPropVariant.bstrVal) : "";

//		m_anArchiveHandler->GetProperty(m_aItemPos, kpidCreationTime, &aPropVariant);
//		m_ftLastCreateTime = aPropVariant.filetime;
//		m_anArchiveHandler->GetProperty(m_aItemPos, kpidLastAccessTime, &aPropVariant);
//		m_ftLastAccessTime = aPropVariant.filetime;
	}
	
	if (lpSubInfo)
	{
		lpSubInfo->uFlag = 0;
		lpSubInfo->uOSType = 0;
		lpSubInfo->dwCRC = m_dwCRC;
		lpSubInfo->dwOriginalSize = (DWORD)m_ui64OriginalSize;
		lpSubInfo->dwCompressedSize = (DWORD)m_ui64CompressedSize;
		lpSubInfo->wRatio = (m_ui64OriginalSize && !m_bSolid) ? (WORD)(m_ui64CompressedSize * 1000 / m_ui64OriginalSize) : 0;
		FILETIME ftLocal;
		::FileTimeToLocalFileTime(&m_ftLastWriteTime, &ftLocal);
		::FileTimeToDosDateTime(&ftLocal, &lpSubInfo->wDate, &lpSubInfo->wTime);
		::lstrcpynA(lpSubInfo->szFileName, m_aFileName, FNAME_MAX32 + 1);
		::lstrcpynA(lpSubInfo->szMode, m_aMode, 8);
		lpSubInfo->szAttribute[0] = (m_dwAttribute & FILE_ATTRIBUTE_ARCHIVE)  ? 'A' : '-';
		lpSubInfo->szAttribute[1] = (m_dwAttribute & FILE_ATTRIBUTE_SYSTEM)   ? 'S' : '-';
		lpSubInfo->szAttribute[2] = (m_dwAttribute & FILE_ATTRIBUTE_HIDDEN)   ? 'H' : '-';
		lpSubInfo->szAttribute[3] = (m_dwAttribute & FILE_ATTRIBUTE_READONLY) ? 'R' : 'W';
		if (m_nArchiveType == ARCHIVETYPE_ZIP)
		{
			lpSubInfo->szAttribute[4] = (m_dwAttribute & FA_ENCRYPTED) ? 'G' : '-';
			lpSubInfo->szAttribute[5] = '\0';
		}
		else
		{
			lpSubInfo->szAttribute[4] = '\0';
			lpSubInfo->szMode[strcspn(lpSubInfo->szMode, " :")] = '\0';
		}
	}

	m_ui64ArcOriginalSize += m_ui64OriginalSize;
	m_ui64ArcCompressedSize += m_ui64CompressedSize;
	++m_aItemPos;

	return 0;
}

int COpenArchive::GetArcFileName(LPSTR lpBuffer, const int nSize)
{
	::lstrcpynA(lpBuffer, GetOemString(m_fdArchive.cFileName), nSize);
	return 0;
}

DWORD COpenArchive::GetArcFileSize()
{
	return m_fdArchive.nFileSizeLow;
}

DWORD COpenArchive::GetArcOriginalSize()
{
	return (DWORD)m_ui64ArcOriginalSize;
}

DWORD COpenArchive::GetArcCompressedSize()
{
	return m_bSolid ? -1 : (DWORD)m_ui64ArcCompressedSize;
}

WORD COpenArchive::GetArcRatio()
{
	if (m_ui64ArcOriginalSize == 0)
		return 0;
	return (WORD)(m_ui64ArcCompressedSize * 1000 / m_ui64ArcOriginalSize);
}

WORD COpenArchive::GetArcDate()
{
	WORD wFatDate;
	WORD wFatTime;
	FILETIME ftLocalFileTime;
	::FileTimeToLocalFileTime(&m_fdArchive.ftLastWriteTime, &ftLocalFileTime);
	::FileTimeToDosDateTime(&ftLocalFileTime, &wFatDate, &wFatTime);
	return wFatDate;
}

WORD COpenArchive::GetArcTime()
{
	WORD wFatDate;
	WORD wFatTime;
	FILETIME ftLocalFileTime;
	::FileTimeToLocalFileTime(&m_fdArchive.ftLastWriteTime, &ftLocalFileTime);
	::FileTimeToDosDateTime(&ftLocalFileTime, &wFatDate, &wFatTime);
	return wFatTime;
}

UINT COpenArchive::GetArcOSType()
{
	return 0;
}

int COpenArchive::IsSFXFile()
{
	return 0;
}

int COpenArchive::GetFileName(LPSTR lpBuffer, const int nSize)
{
	if (!m_bSearchMode)
		return ERROR_NOT_SEARCH_MODE;
	::lstrcpynA(lpBuffer, m_aFileName, nSize);
	return 0;
}

DWORD COpenArchive::GetOriginalSize()
{
	if (!m_bSearchMode)
		return -1;
	return (DWORD)m_ui64OriginalSize;
}

DWORD COpenArchive::GetCompressedSize()
{
	if (!m_bSearchMode)
		return -1;
	return (DWORD)m_ui64CompressedSize;
}

WORD COpenArchive::GetRatio()
{
	if (!m_bSearchMode)
		return -1;
	if (m_ui64OriginalSize == 0)
		return 0;
	return (WORD)(m_ui64CompressedSize * 1000 / m_ui64OriginalSize);
}

WORD COpenArchive::GetDate()
{
	if (!m_bSearchMode)
		return -1;
	WORD wFatDate;
	WORD wFatTime;
	FILETIME ftLocalFileTime;
	::FileTimeToLocalFileTime(&m_ftLastWriteTime, &ftLocalFileTime);
	::FileTimeToDosDateTime(&ftLocalFileTime, &wFatDate, &wFatTime);
	return wFatDate;
}

WORD COpenArchive::GetTime()
{
	if (!m_bSearchMode)
		return -1;
	WORD wFatDate;
	WORD wFatTime;
	FILETIME ftLocalFileTime;
	::FileTimeToLocalFileTime(&m_ftLastWriteTime, &ftLocalFileTime);
	::FileTimeToDosDateTime(&ftLocalFileTime, &wFatDate, &wFatTime);
	return wFatTime;
}

DWORD COpenArchive::GetCRC()
{
	if (!m_bSearchMode)
		return -1;
	return m_dwCRC;
}

int COpenArchive::GetAttribute()
{
	if (!m_bSearchMode)
		return -1;
	return m_dwAttribute & 0xff;
}

UINT COpenArchive::GetOSType()
{
	if (!m_bSearchMode)
		return -1;
	return 0;
}

int COpenArchive::GetMethod(LPSTR lpBuffer, const int nSize)
{
	if (!m_bSearchMode)
		return ERROR_NOT_SEARCH_MODE;
	::lstrcpynA(lpBuffer, m_aMode, nSize);
	if (nSize == 8 && m_nArchiveType == ARCHIVETYPE_7Z)
		lpBuffer[strcspn(lpBuffer, " :")] = '\0';
	return 0;
}

DWORD COpenArchive::GetWriteTime()
{
	if (!m_bSearchMode)
		return -1;
	struct tm atm;
	FILETIME ftLocalFileTime;
	SYSTEMTIME stLastWriteTime;
	::FileTimeToLocalFileTime(&m_ftLastWriteTime, &ftLocalFileTime);
	::FileTimeToSystemTime(&ftLocalFileTime, &stLastWriteTime);
	atm.tm_year = stLastWriteTime.wYear - 1900;
	atm.tm_mon = stLastWriteTime.wMonth - 1;
	atm.tm_mday = stLastWriteTime.wDay;
	atm.tm_hour = stLastWriteTime.wHour;
	atm.tm_min = stLastWriteTime.wMinute;
	atm.tm_sec = stLastWriteTime.wSecond;
	atm.tm_isdst = -1;
	return mktime(&atm);
}
/*
DWORD COpenArchive::GetCreateTime()
{
	if (!m_bSearchMode || m_nArchiveType == ARCHIVETYPE_ZIP)
		return -1;
	struct tm atm;
	FILETIME ftLocalFileTime;
	SYSTEMTIME stLastCreateTime;
	::FileTimeToLocalFileTime(&m_ftLastCreateTime, &ftLocalFileTime);
	::FileTimeToSystemTime(&ftLocalFileTime, &stLastCreateTime);
	atm.tm_year = stLastCreateTime.wYear - 1900;
	atm.tm_mon = stLastCreateTime.wMonth - 1;
	atm.tm_mday = stLastCreateTime.wDay;
	atm.tm_hour = stLastCreateTime.wHour;
	atm.tm_min = stLastCreateTime.wMinute;
	atm.tm_sec = stLastCreateTime.wSecond;
	atm.tm_isdst = -1;
	return mktime(&atm);
}

DWORD COpenArchive::GetAccessTime()
{
	if (!m_bSearchMode || m_nArchiveType == ARCHIVETYPE_ZIP)
		return -1;
	struct tm atm;
	FILETIME ftLocalFileTime;
	SYSTEMTIME stLastAccessTime;
	::FileTimeToLocalFileTime(&m_ftLastAccessTime, &ftLocalFileTime);
	::FileTimeToSystemTime(&ftLocalFileTime, &stLastAccessTime);
	atm.tm_year = stLastAccessTime.wYear - 1900;
	atm.tm_mon = stLastAccessTime.wMonth - 1;
	atm.tm_mday = stLastAccessTime.wDay;
	atm.tm_hour = stLastAccessTime.wHour;
	atm.tm_min = stLastAccessTime.wMinute;
	atm.tm_sec = stLastAccessTime.wSecond;
	atm.tm_isdst = -1;
	return mktime(&atm);
}
*/
BOOL COpenArchive::GetWriteTimeEx(FILETIME *lpftLastWriteTime)
{
	if (!m_bSearchMode || m_nArchiveType == ARCHIVETYPE_ZIP)
		return FALSE;
	*lpftLastWriteTime = m_ftLastWriteTime;
	return TRUE;
}
/*
BOOL COpenArchive::GetCreateTimeEx(FILETIME *lpftLastCreateTime)
{
	if (!m_bSearchMode || m_nArchiveType == ARCHIVETYPE_ZIP)
		return FALSE;
	*lpftLastCreateTime = m_ftLastCreateTime;
	return TRUE;
}

BOOL COpenArchive::GetAccessTimeEx(FILETIME *lpftLastAccessTime)
{
	if (!m_bSearchMode || m_nArchiveType == ARCHIVETYPE_ZIP)
		return FALSE;
	*lpftLastAccessTime = m_ftLastAccessTime;
	return TRUE;
}
*/
BOOL COpenArchive::GetArcCreateTimeEx(FILETIME *lpftCreationTime)
{
	*lpftCreationTime = m_fdArchive.ftCreationTime;
	return TRUE;
}

BOOL COpenArchive::GetArcAccessTimeEx(FILETIME *lpftLastAccessTime)
{
	*lpftLastAccessTime = m_fdArchive.ftLastAccessTime;
	return TRUE;
}

BOOL COpenArchive::GetArcWriteTimeEx(FILETIME *lpftLastWriteTime)
{
	*lpftLastWriteTime = m_fdArchive.ftLastWriteTime;
	return TRUE;
}

BOOL COpenArchive::GetArcFileSizeEx(ULHA_INT64* lpllSize)
{
	*lpllSize = ((ULHA_INT64)m_fdArchive.nFileSizeHigh << 32) + m_fdArchive.nFileSizeLow;
	return TRUE;
}

BOOL COpenArchive::GetArcOriginalSizeEx(ULHA_INT64* lpllSize)
{
	*lpllSize = (ULHA_INT64)m_ui64ArcOriginalSize;
	return TRUE;
}

BOOL COpenArchive::GetArcCompressedSizeEx(ULHA_INT64* lpllSize)
{
	*lpllSize = m_bSolid ? -1 : (ULHA_INT64)m_ui64ArcCompressedSize;
	return TRUE;
}

BOOL COpenArchive::GetOriginalSizeEx(ULHA_INT64* lpllSize)
{
	if (!m_bSearchMode)
		return FALSE;
	*lpllSize = (ULHA_INT64)m_ui64OriginalSize;
	return TRUE;
}

BOOL COpenArchive::GetCompressedSizeEx(ULHA_INT64* lpllSize)
{
	if (!m_bSearchMode)
		return FALSE;
	*lpllSize = (ULHA_INT64)m_ui64CompressedSize;
	return TRUE;
}

int COpenArchive::GetFileCount()
{
	return m_aNumItems;
}
